--[[
    Aaron MacDonald
    August 18, 2010
]]

require('graphics')
require('constants')
require('actor')
require('levels')
require('ai')

local FPS = 60
local SECONDS_PER_FRAME = 1 / FPS

local accumulatedTime

local isPaused

----------------------------------------------------------------------

local KEY_QUIT = 'escape'

local VALID_BUTTONS = { [constants.UP] = true,
                        [constants.DOWN] = true,
                        [constants.LEFT] = true,
                        [constants.RIGHT] = true }

----------------------------------------------------------------------

local PLAYER_TYPE = actor.buildType({
    spriteSheet = 'player',
    spriteWidth = 64, spriteHeight = 96,
    offsetY = 64,
    groundRadius = 32,
    bodyWidth = 40,
    bodyHeight = 64,
    speed = 600 -- I have no freakin' clue what units this is in
})

local NPC_TYPE = actor.buildType({
    spriteSheet = 'npc',
    spriteWidth = 64, spriteHeight = 96,
    offsetY = 64,
    groundRadius = 32,
    bodyWidth = 40,
    bodyHeight = 64,
    speed = 600
})

----------------------------------------------------------------------

local currentButtons = { [constants.UP] = false,
                         [constants.DOWN] = false,
                         [constants.LEFT] = false,
                         [constants.RIGHT] = false }

local currentLevel
local allActors
local player

local camera

local START_LEVEL = 'test_level'

local newlyTeleported

----------------------------------------------------------------------

local wanderAISteering = {
    [constants.NORTH] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
        end,
    [constants.NORTHEAST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.UP] = true
            movement[constants.RIGHT] = true
        end,
    [constants.EAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.LEFT] = false

            movement[constants.RIGHT] = true
        end,
    [constants.SOUTHEAST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false

            movement[constants.DOWN] = true
            movement[constants.RIGHT] = true
        end,
    [constants.SOUTH] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.LEFT] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
        end,
    [constants.SOUTHWEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.RIGHT] = false

            movement[constants.DOWN] = true
            movement[constants.LEFT] = true
        end,
    [constants.WEST] =
        function(movement)
            movement[constants.UP] = false
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.LEFT] = true
        end,
    [constants.NORTHWEST] =
        function(movement)
            movement[constants.DOWN] = false
            movement[constants.RIGHT] = false

            movement[constants.UP] = true
            movement[constants.LEFT] = true
        end,
}

local function wanderAISelectDir(brain)
    local i = math.random(1, #brain.possibleDirs)
    dir = brain.possibleDirs[i]
    wanderAISteering[dir](brain.movement)
end

local function wanderAIInit(brain)
    brain.possibleDirs = {constants.NORTH, constants.NORTHEAST,
      constants.EAST, constants.SOUTHEAST, constants.SOUTH,
      constants.SOUTHWEST, constants.WEST, constants.NORTHWEST}

    wanderAISelectDir(brain)

    brain.travelTime = 10
end

local function wanderAIThink(brain)
    if brain.travelTime == 0 then
        wanderAISelectDir(brain)
        brain.travelTime = 10
    else
        brain.travelTime = brain.travelTime - 1
    end
end

--[[
local function wanderHandleWallHit(brain)
    wanderAISelectDir(brain)
    brain.travelTime = 10
end
]]

----------------------------------------------------------------------

local function spawnNPCs(level)
    local newNPCs = {}

    if level.npcSpawn then
        for _, data in ipairs(level.npcSpawn) do
            local npc = actor.newActor(level, NPC_TYPE,
              data.x, data.y)

            npc.ai = ai.newBrain(5, wanderAIThink, wanderAIInit)--,
              --wanderHandleWallHit)

            table.insert(newNPCs, npc)
        end
    end

    return newNPCs
end

----------------------------------------------------------------------

local function setCamera()
    if levels.pixelWidth(currentLevel) < camera.width
    then
        camera.scrollX = false
        camera.x = -(camera.width - levels.pixelWidth(currentLevel))
          / 2
    else
        camera.scrollX = true

        camera.minScrollX = camera.width / 2
        camera.maxScrollX = levels.pixelWidth(currentLevel)
          - camera.width
    end

    if levels.pixelHeight(currentLevel) < camera.height
    then
        camera.scrollY = false
        camera.y = currentLevel.minY - (
          (camera.height - levels.truePixelHeight(currentLevel)) / 2)
    else
        camera.scrollY = true

        camera.minScrollY = camera.height / 2
        camera.maxScrollY = levels.pixelHeight(currentLevel)
          - camera.height
    end
end

----------------------------------------------------------------------

-- local ACTION_SEP = ', '
-- 
-- local HIT_ACTIONS = { }
-- local UNHIT_ACTIONS = { }
-- 
-- local function addCollision(a, b, coll)
--     if a and b then
--         local action = a.type .. ACTION_SEP .. b.type
-- 
--         if HIT_ACTIONS[action] then
--             HIT_ACTIONS[action](a.data, b.data, coll)
--         else
--             action = b.type .. ACTION_SEP .. a.type
-- 
--             if HIT_ACTIONS[action] then
--                 HIT_ACTIONS[action](b.data, a.data, coll)
--             end
--         end
--     end
-- end
-- 
-- local function removeCollision(a, b, coll)
--     if a and b then
--         local action = a.type .. ACTION_SEP .. b.type
-- 
--         if UNHIT_ACTIONS[action] then
--             UNHIT_ACTIONS[action](a.data, b.data, coll)
--         else
--             action = b.type .. ACTION_SEP .. a.type
-- 
--             if UNHIT_ACTIONS[action] then
--                 UNHIT_ACTIONS[action](b.data, a.data, coll)
--             end
--         end
--     end
-- end
-- 
-- -- Darn cyclic dependencies
-- HIT_ACTIONS = {
-- 
-- }
-- 
-- UNHIT_ACTIONS = {
-- 
-- }

----------------------------------------------------------------------

function love.load()
    math.randomseed( os.time() )

    accumulatedTime = 0
    isPaused = false

    graphics.loadFont()

    --

    currentLevel = levels.newLevel(START_LEVEL)

    camera = {
        x = 0,
        y = 0,
        width = love.graphics.getWidth(),
        height = love.graphics.getHeight()
    }
    setCamera()

    --

    local startX = 0
    local startY = 0

    if currentLevel.playerSpawn then
        startX = currentLevel.playerSpawn.x
        startY = currentLevel.playerSpawn.y
    end

    player = actor.newActor(currentLevel, PLAYER_TYPE, startX, startY,
      true)

    allActors = spawnNPCs(currentLevel)
    table.insert(allActors, player)

    newlyTeleported = false
end

----------------------------------------------------------------------

function love.keypressed( key, unicode )
    if key == KEY_QUIT then
        love.event.push('q')
    elseif VALID_BUTTONS[key] then
        currentButtons[key] = true
    end
end

function love.keyreleased( key, unicode )
   if VALID_BUTTONS[key] then
       currentButtons[key] = false
   end
end

function love.focus(hasFocus)
    isPaused = not hasFocus
end

----------------------------------------------------------------------

local function updateCamera()
    if camera.scrollX
    then
        if actor.getRoundedX(player) < camera.minScrollX then
            camera.x = 0
        elseif actor.getRoundedX(player)
          > (camera.maxScrollX + camera.minScrollX)
        then
            camera.x = camera.maxScrollX
        else
            camera.x = actor.getRoundedX(player) - camera.minScrollX
        end
    end

    if camera.scrollY
    then
        if (actor.getRoundedY(player) - currentLevel.minY)
          < camera.minScrollY
        then
            camera.y = currentLevel.minY
        elseif actor.getRoundedY(player)
          > (camera.maxScrollY + camera.minScrollY)
        then
            camera.y = camera.maxScrollY
        else
            camera.y = actor.getRoundedY(player) - camera.minScrollY
        end
    end
end

local function doUpdate()
    -- Update velocities

    actor.updateMotion(player, currentButtons)

    local movementControls
    for _, act in ipairs(allActors) do
        if act == player then
            movementControls = currentButtons
        else
            ai.think(act.ai)
            movementControls = act.ai.movement
        end

        actor.updateMotion(act, movementControls)

        if act ~= player then
            ai.update(act.ai)
        end
    end

    --

    currentLevel.physicsWorld:update(SECONDS_PER_FRAME)

    local teleport = levels.exitData(currentLevel,
      actor.getX(player), actor.getY(player))

    if teleport then
        if not newlyTeleported then
            currentLevel = levels.newLevel(teleport.level)

            setCamera()

            x, y = levels.entranceCoords(currentLevel, teleport.dest)

            actor.changeLevel(player, currentLevel, x, y, true)

            allActors = spawnNPCs(currentLevel)
            table.insert(allActors, player)

            newlyTeleported = true
        end
    else
        newlyTeleported = false
    end

    updateCamera()
end

function love.update(dt)
    accumulatedTime = accumulatedTime + dt
    while accumulatedTime >= SECONDS_PER_FRAME
    do
        if not isPaused then
            doUpdate()
        end
        accumulatedTime = accumulatedTime - SECONDS_PER_FRAME
    end
end

----------------------------------------------------------------------

local function spriteCompare(a, b)
    local ay = actor.getY(a)
      + (a.type.spriteHeight - a.type.spriteOffsetY)
    local by = actor.getY(b)
      + (b.type.spriteHeight - b.type.spriteOffsetY)

    return ay < by
end

function love.draw()
    love.graphics.translate(-camera.x, -camera.y)

    levels.drawGround(currentLevel, camera)

    table.sort(allActors, spriteCompare)

    for _, act in ipairs(allActors) do
        actor.draw(act)
    end

    levels.drawOver(currentLevel, camera)

    love.graphics.translate(camera.x, camera.y)
end